/*
 * settings.h - Definitions of settings keys and schema.
 */

#ifndef __SETTINGS_H__
#define __SETTINGS_H__

#define NOTIFICATIONS_SCHEMA              "net.launchpad.indicator.notifications"
#define NOTIFICATIONS_KEY_BLACKLIST       "blacklist"
#define NOTIFICATIONS_KEY_BLACKLIST_HINTS "blacklist-hints"
#define NOTIFICATIONS_KEY_CLEAR_MC        "clear-on-middle-click"
#define NOTIFICATIONS_KEY_DND             "do-not-disturb"
#define NOTIFICATIONS_KEY_HIDE_INDICATOR  "hide-indicator"
#define NOTIFICATIONS_KEY_MAX_ITEMS       "max-items"

#define MATE_SCHEMA  "org.mate.NotificationDaemon"
#define MATE_KEY_DND "do-not-disturb"

#endif
